<?php

require_once('NodeRPC_Service_Exception.php');

abstract class NodeRPC_Service_Abstract {

    /**
     * @var string
     */
    private $host;

    /**
     * @var int
     */
    private $port;

    /**
     * @var stdClass
     */
    protected $serviceDefinition;

    /**
     * @var string
     */
    protected $serviceName;

    public function __construct($host, $port = 80) {
        $this->host = $host;
        $this->port = $port;
        $this->init();
    }

    /**
     * Initialize service
     *
     * @abstract
     * @return void
     */
    abstract function init();

    /**
     * Convert arguments to http format
     *
     * @param  $args
     * @param  $paramDefinition
     * @return void
     */
    protected function marshalParams($args, $paramDefinition) {
        $params = array();
        foreach ($args as $i => &$arg) {
            $params[$paramDefinition[$i]['name']] = $arg;
        }
        return $params;
    }

    /**
     * Make http request with HttpRequest Object
     *
     * @throws NodeRPC_Service_Exception
     * @param  $url
     * @return mixed
     */
    private function httpHttpRequest($url, $params) {
        $httpRequest = new HttpRequest($url, HttpRequest::METH_POST);
        $httpRequest->addHeaders(array(
                                      'Content-type' => 'application/json'
                                 ));
        $httpRequest->setBody(json_encode($params));
        $httpRequest->send();

        if ($httpRequest->getResponseCode() == 200) {
            return json_decode($httpRequest->getResponseBody());
        }
        else {
            throw new NodeRPC_Service_Exception($httpRequest->getResponseBody());
        }
    }

    /**
     * Make http request with fsockopen
     *
     * @throws Exception
     * @param  $url
     * @return mixed
     */
    private function httpFsockOpen($url, $params) {
        $fp = fsockopen($this->host, $this->port, $errno, $errstr, $timeout = 30);
        if (!$fp) {
            throw new NodeRPC_Service_Exception($errstr, $errno);
        }
        else {
            $str = json_encode($params);
            fputs($fp, "POST $url HTTP/1.1\r\n");
            fputs($fp, "Host: {$this->host}:{$this->port}\r\n");
            fputs($fp, "Accept: */*\r\n");
            fputs($fp, "Content-type: application/json\r\n");
            fputs($fp, "Content-length: " . mb_strlen($str, 'ascii') . "\r\n");
            fputs($fp, "\r\n");
            fputs($fp, $str);

            $response = "";
            while (!feof($fp)) {
                $response .= fgets($fp, 4096);
                list($header, $body) = explode("\r\n\r\n", $response);
                if ($body > '') {
                    list($contentLength, $body) = explode("\r\n", $body, 2);
                    $contentLength = hexdec($contentLength);
                    if ($contentLength == 0) {
                        break;
                    }

                    $body = substr($body, 0, $contentLength);
                    if (mb_strlen($body, 'ascii') == $contentLength) {
                        break;
                    }
                }
            }
            fclose($fp);
        }

        if (preg_match("^HTTP.* 200 OK", $header)) {
            return json_decode($body);
        }
        else {
            throw new NodeRPC_Service_Exception($body);
        }
    }

    /**
     * Send request via http
     * @param  $method
     * @param  $args
     * @return void
     */
    protected function proxyCall($method, $args) {
        $params = $this->marshalParams($args, $this->serviceDefinition[$method]['params']);
        $route = "http://" . $this->host . ':' . $this->port . '/' . $this->serviceName . "/" . $method;
        if (class_exists('HttpRequest')) {
            return $this->httpHttpRequest($route, $params);
        }
        else {
            return $this->httpFsockOpen($route, $params);
        }
    }

}